/* mbed NokiaLCD Library, for a 130x130 Nokia colour LCD
 * Copyright (c) 2007-2010, sford
 *
 * Permission is hereby granted, free of charge, to any person obtaining a copy
 * of this software and associated documentation files (the "Software"), to deal
 * in the Software without restriction, including without limitation the rights
 * to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
 * copies of the Software, and to permit persons to whom the Software is
 * furnished to do so, subject to the following conditions:
 *
 * The above copyright notice and this permission notice shall be included in
 * all copies or substantial portions of the Software.
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
 * AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
 * LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
 * OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
 * THE SOFTWARE.
 */

#ifndef MBED_NOKIALCD_H
#define MBED_NOKIALCD_H

#include  <inttypes.h>
#include "Print.h"

#define NOKIA_PORT     PORTA
#define NOKIA_DDR      DDRA
#define NOKIA_MOSI     4
#define NOKIA_SCLK     2
#define NOKIA_CS       3
#define NOKIA_RST      1

#define NOKIA_MOSI_0 cbi(NOKIA_PORT, NOKIA_MOSI)
#define NOKIA_MOSI_1 sbi(NOKIA_PORT, NOKIA_MOSI)

#define NOKIA_SCLK_0 cbi(NOKIA_PORT, NOKIA_SCLK)
#define NOKIA_SCLK_1 sbi(NOKIA_PORT, NOKIA_SCLK)

#define NOKIA_CS_0 cbi(NOKIA_PORT, NOKIA_CS)
#define NOKIA_CS_1 sbi(NOKIA_PORT, NOKIA_CS)

#define NOKIA_RST_0 cbi(NOKIA_PORT, NOKIA_RST)
#define NOKIA_RST_1 sbi(NOKIA_PORT, NOKIA_RST)


/** An interface for the 130x130 Nokia Mobile phone screens
 *
 * @code
 * #include "mbed.h"
 * #include "NokiaLCD.h"
 *
 * NokiaLCD lcd(p5, p7, p8, p9); // mosi, sclk, cs, rst
 *
 * int main() {
 *     lcd.printf("Hello World!");
 * }
 */
class NokiaLCD : public Print 
{

public:
    /** LCD panel format */
    enum LCDType {
        LCD6100     /**< Nokia 6100, as found on sparkfun board (default) */
        , LCD6610   /**< Nokia 6610, as found on olimex board */
        , PCF8833
    };

    /** Create and Nokia LCD interface, using a SPI and two DigitalOut interfaces
     *
     * @param mosi SPI data out
     * @param sclk SPI clock
     * @param cs Chip Select (DigitalOut)
     * @param rst Reset (DigitalOut)
     * @param type The LCDType to select driver chip variants
     */
//    NokiaLCD(uint8_t mosi, uint8_t sclk, uint8_t cs, uint8_t rst, LCDType type = LCD6100);
      NokiaLCD(uint8_t mosi, uint8_t sclk, uint8_t cs, uint8_t rst);

#if DOXYGEN_ONLY
    /** Write a character to the LCD
     *
     * @param c The character to write to the display
     */
    int putc(int c);

    /** Write a formated string to the LCD
     *
     * @param format A printf-style format string, followed by the
     *               variables to use in formating the string.
     */
    int printf(const char* format, ...);
#endif

    /** Locate to a screen column and row
     *
     * @param column  The horizontal position from the left, indexed from 0
     * @param row     The vertical position from the top, indexed from 0
     */
    void locate(int column, int row);

    /** Clear the screen and locate to 0,0 */
    void cls();

    void pixel(int x, int y, int colour);
    void fill(int x, int y, int width, int height, int colour);
    void circle(int x0, int y0, int r, int colour);                 // Added this to create circles easy. 
    void line(int x0, int y0, int x1, int y1, int colour);          // Added this to create lines easy. 

    void blit(int x, int y, int width, int height, const int* colour);
    void bitblit(int x, int y, int width, int height, const char* bitstream);

    int width();
    int height();
    int columns();
    int rows();

    void reset();

    void foreground(int c);
    void background(int c);

    virtual void write(uint8_t);

protected:
    virtual void _window(int x, int y, int width, int height);
    virtual void _putp(int colour);

    void command(uint8_t value);
    void data(uint8_t value);

    void newline();
    virtual int _putc(int c);
    virtual int _getc() {
        return 0;
    }
    void putp(int v);
    void window(int x, int y, int width, int height);

    LCDType _type;
    int _row, _column, _rows, _columns, _foreground, _background, _width, _height;
};

#endif


